@@ -18,30 +18,41 @@
 namespace memray::tracking_api {
 
 extern const char MAGIC[7];  // Value assigned in records.cpp
-const int CURRENT_HEADER_VERSION = 11;
+const int CURRENT_HEADER_VERSION = 12;
 
 using frame_id_t = size_t;
 using thread_id_t = unsigned long;
 using millis_t = long long;
+using code_object_id_t = size_t;
 
+// If the high (128) bit is set on a given record type discriminator,
+// it's an ALLOCATION record with 7 bits available for flags (see
+// "ALLOCATION ENCODING" in record_writer.cpp for details).
+//
+// Otherwise, if the 64 bit is set, it's a FRAME_PUSH record with 6 bits
+// available for flags (see "FRAME_PUSH ENCODING" in record_writer.cpp).
+//
+// Otherwise, if the 16 bit is set, it's a FRAME_POP record with 4 bits
+// available for flags (see "FRAME_POP ENCODING" in record_writer.cpp).
+//
+// Otherwise, it's a record type that has no flags, and all remaining
+// bits identify the record type
 enum class RecordType : unsigned char {
-    OTHER = 0,
-    ALLOCATION = 1,
-    ALLOCATION_WITH_NATIVE = 2,
-    FRAME_INDEX = 3,
-    FRAME_PUSH = 4,
+    FILLER = 0,
+    TRAILER = 1,
+    MEMORY_RECORD = 2,
     NATIVE_TRACE_INDEX = 5,
     MEMORY_MAP_START = 6,
     SEGMENT_HEADER = 7,
     SEGMENT = 8,
-    FRAME_POP = 9,
     THREAD_RECORD = 10,
-    MEMORY_RECORD = 11,
     CONTEXT_SWITCH = 12,
-};
+    CODE_OBJECT = 14,
 
-enum class OtherRecordType : unsigned char {
-    TRAILER = 1,
+    FRAME_POP = 16,  // 16 through 31
+    OBJECT_RECORD = 32,  // 32 through 63
+    FRAME_PUSH = 64,  // 64 through 127
+    ALLOCATION = 128,  // 128 through 255
 };
 
 // Enumerators that have the same name as in RecordType are encoded the same
@@ -59,33 +70,12 @@ enum class AggregatedRecordType : unsigned char {
     SEGMENT = 8,
     THREAD_RECORD = 10,
     CONTEXT_SWITCH = 12,
+    SURVIVING_OBJECT = 13,
+    CODE_OBJECT = 14,
 
     AGGREGATED_TRAILER = 15,
 };
 
-struct RecordTypeAndFlags
-{
-    RecordTypeAndFlags()
-    : record_type(RecordType::OTHER)
-    , flags(0)
-    {
-    }
-
-    RecordTypeAndFlags(RecordType record_type_, unsigned char flags_)
-    : record_type(record_type_)
-    , flags(flags_)
-    {
-        // Ensure both values fit into 4 bits
-        assert(static_cast<int>(record_type_) == (static_cast<int>(record_type_) & 0x0f));
-        assert(static_cast<int>(flags_) == (static_cast<int>(flags_) & 0x0f));
-    }
-
-    RecordType record_type : 4;
-    unsigned char flags : 4;
-};
-
-static_assert(sizeof(RecordTypeAndFlags) == 1);
-
 struct TrackerStats
 {
     size_t n_allocations{0};
@@ -112,6 +102,7 @@ struct HeaderRecord
 {
     char magic[sizeof(MAGIC)];
     int version{};
+    int python_version{PY_VERSION_HEX};
     bool native_traces{false};
     FileFormat file_format{FileFormat::ALL_ALLOCATIONS};
     TrackerStats stats{};
@@ -121,29 +112,23 @@ struct HeaderRecord
     size_t skipped_frames_on_main_tid{};
     PythonAllocatorType python_allocator{};
     bool trace_python_allocators{};
+    bool track_object_lifetimes{false};
 };
 
 struct MemoryRecord
 {
-    unsigned long int ms_since_epoch;
+    uint64_t ms_since_epoch;
     size_t rss;
 };
 
 struct MemorySnapshot
 {
-    unsigned long int ms_since_epoch;
+    uint64_t ms_since_epoch;
     size_t rss;
     size_t heap;
 };
 
 struct AllocationRecord
-{
-    uintptr_t address;
-    size_t size;
-    hooks::Allocator allocator;
-};
-
-struct NativeAllocationRecord
 {
     uintptr_t address;
     size_t size;
@@ -165,6 +150,18 @@ struct Allocation
     PyObject* toPythonObject() const;
 };
 
+struct TrackedObject
+{
+    thread_id_t tid;
+    uintptr_t address;
+    bool is_created;
+    frame_id_t native_frame_id{0};
+    size_t frame_index{0};
+    size_t native_segment_generation{0};
+
+    PyObject* toPythonObject() const;
+};
+
 struct AggregatedAllocation
 {
     thread_id_t tid;
@@ -206,79 +203,86 @@ struct ImageSegments
     std::vector<Segment> segments;
 };
 
-struct RawFrame
-{
-    const char* function_name;
-    const char* filename;
-    int lineno;
-    bool is_entry_frame;
-
-    auto operator==(const RawFrame& other) const -> bool
-    {
-        return (function_name == other.function_name && filename == other.filename
-                && lineno == other.lineno && is_entry_frame == other.is_entry_frame);
-    }
-
-    struct Hash
-    {
-        auto operator()(memray::tracking_api::RawFrame const& frame) const noexcept -> std::size_t
-        {
-            // Keep this hashing fast and simple as this has a non trivial
-            // performance impact on the tracing functionality. We don't hash
-            // the contents of the strings because the interpreter will give us
-            // the same char* for the same code object. Of course, we can have
-            // some scenarios where two functions with the same function name have
-            // two different char* but in that case we will end registering the
-            // name twice, which is a good compromise given the speed that we
-            // gain keeping this simple.
-
-            auto the_func = std::hash<const char*>{}(frame.function_name);
-            auto the_filename = std::hash<const char*>{}(frame.filename);
-            auto lineno = std::hash<int>{}(frame.lineno);
-            return the_func ^ the_filename ^ lineno ^ frame.is_entry_frame;
-        }
-    };
-};
-
 struct Frame
 {
-    std::string function_name;
-    std::string filename;
-    int lineno{0};
-    bool is_entry_frame{true};
-
-    PyObject* toPythonObject(python_helpers::PyUnicode_Cache& pystring_cache) const;
+    code_object_id_t code_object_id;
+    int instruction_offset;
+    bool is_entry_frame;
 
     auto operator==(const Frame& other) const -> bool
     {
-        return (function_name == other.function_name && filename == other.filename
-                && lineno == other.lineno && is_entry_frame == other.is_entry_frame);
+        return (code_object_id == other.code_object_id && instruction_offset == other.instruction_offset
+                && is_entry_frame == other.is_entry_frame);
     }
 
     struct Hash
     {
         auto operator()(memray::tracking_api::Frame const& frame) const noexcept -> std::size_t
         {
-            // Keep this hashing fast and simple as this has a non trivial
-            // performance impact on the tracing functionality. We don't hash
-            // the contents of the strings because the interpreter will give us
-            // the same char* for the same code object. Of course, we can have
-            // some scenarios where two functions with the same function name have
-            // two different char* but in that case we will end registering the
-            // name twice, which is a good compromise given the speed that we
-            // gain keeping this simple.
-
-            auto the_func = std::hash<std::string>{}(frame.function_name);
-            auto the_filename = std::hash<std::string>{}(frame.filename);
-            auto lineno = std::hash<int>{}(frame.lineno);
-            return the_func ^ the_filename ^ lineno ^ frame.is_entry_frame;
+            return std::hash<uint64_t>()(frame.code_object_id)
+                   ^ std::hash<int>()(frame.instruction_offset) ^ frame.is_entry_frame;
         }
     };
 };
 
+struct Location
+{
+    std::string function_name;
+    std::string filename;
+    int lineno{0};
+
+    PyObject* toPythonObject(python_helpers::PyUnicode_Cache& pystring_cache) const;
+
+    auto operator==(const Location& other) const -> bool
+    {
+        return (function_name == other.function_name && filename == other.filename
+                && lineno == other.lineno);
+    }
+
+    struct Hash
+    {
+        auto operator()(const Location& location) const noexcept -> std::size_t
+        {
+            auto the_func = std::hash<std::string>{}(location.function_name);
+            auto the_filename = std::hash<std::string>{}(location.filename);
+            auto lineno = std::hash<int>{}(location.lineno);
+            return the_func ^ the_filename ^ lineno;
+        }
+    };
+};
+
+// For storing code object info with strings (used in reader)
+struct CodeObjectInfo
+{
+    std::string function_name;
+    std::string filename;
+    std::string linetable;
+    int firstlineno;
+};
+
+using pycode_map_val_t = std::pair<code_object_id_t, CodeObjectInfo>;
+
+// Structure to represent code object information
+struct CodeObject
+{
+    const char* function_name;
+    const char* filename;
+    const char* linetable;
+    size_t linetable_size;
+    int firstlineno;
+};
+
+struct RawFrame
+{
+    PyCodeObject* code;
+    CodeObject code_info;
+    bool is_entry_frame;
+    int instruction_offset;
+};
+
 struct FramePush
 {
-    frame_id_t frame_id;
+    Frame frame;
 };
 
 struct FramePop
@@ -303,40 +307,54 @@ struct DeltaEncodedFields
     uintptr_t instruction_pointer{};
     uintptr_t data_pointer{};
     frame_id_t native_frame_id{};
-    frame_id_t python_frame_id{};
-    int python_line_number{};
+    int code_firstlineno{};
 };
 
-template<typename FrameType>
-class FrameCollection
+template<typename RecordType>
+class Registry
 {
   public:
-    template<typename T>
-    auto getIndex(T&& frame) -> std::pair<frame_id_t, bool>
+    using index_t = size_t;
+
+    size_t size() const
     {
-        bool inserted = false;
-        auto it = d_frame_map.find(frame);
-        if (it == d_frame_map.end()) {
-            std::tie(it, inserted) = d_frame_map.emplace(std::forward<T>(frame), d_current_frame_id);
-            if (inserted) {
-                d_current_frame_id++;
-            }
+        return d_record_by_id.size();
+    }
+
+    std::pair<index_t, bool> registerRecord(const RecordType& record)
+    {
+        auto [it, inserted] = d_id_by_record.emplace(record, d_record_by_id.size());
+        if (inserted) {
+            d_record_by_id.push_back(record);
         }
         return std::make_pair(it->second, inserted);
     }
 
-  private:
-    frame_id_t d_current_frame_id{};
-    std::unordered_map<FrameType, frame_id_t, typename FrameType::Hash> d_frame_map{};
-};
+    RecordType& getRecord(index_t index)
+    {
+        return d_record_by_id[index];
+    }
 
-using pyrawframe_map_val_t = std::pair<frame_id_t, RawFrame>;
-using pyframe_map_val_t = std::pair<frame_id_t, Frame>;
-using pyframe_map_t = std::unordered_map<pyframe_map_val_t::first_type, pyframe_map_val_t::second_type>;
+    const RecordType& getRecord(index_t index) const
+    {
+        return d_record_by_id[index];
+    }
+
+  private:
+    std::unordered_map<RecordType, index_t, typename RecordType::Hash> d_id_by_record{};
+    std::vector<RecordType> d_record_by_id{};
+};
 
 struct ThreadRecord
 {
     const char* name;
 };
 
+struct ObjectRecord
+{
+    uintptr_t address;  // Address of the PyObject*
+    bool is_created;  // true for creation, false for destruction
+    frame_id_t native_frame_id{0};  // Optional native frame id for backtraces
+};
+
 }  // namespace memray::tracking_api
